//---------------------------------------------------------------------------
//
//  SCSI Target Emulator RaSCSI (*^..^*)
//  for Raspberry Pi
//
//  Powered by XM6 TypeG Technorogy.
//  Copyright (C) 2016-2017 GIMONS
//	[ RaSCSI C[T[lbg SCSI䕔 ]
//
//  Based on
//    Neptune-X board driver for  Human-68k(ESP-X)   version 0.03
//    Programed 1996-7 by Shi-MAD.
//    Special thanks to  Niggle, FIRST, yamapu ...
//
//---------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <sys/dos.h>
#include <sys/scsi.h>
#include <iocslib.h>
#include "main.h"
#include "scsictl.h"

typedef struct
{
	char DeviceType;
	char RMB;
	char ANSI_Ver;
	char RDF;
	char AddLen;
	char RESV0;
	char RESV1;
	char OptFunc;
	char VendorID[8];
	char ProductID[16];
	char FirmRev[4];
} INQUIRY_T;

typedef struct
{
	INQUIRY_T info;
	char options[8];
} INQUIRYOPT_T;

#define MFP_AEB		0xe88003
#define MFP_IERB	0xe88009
#define MFP_IMRB	0xe88015

// asmsub.s ̃Tu[`
extern void DI();
extern void EI();

volatile short* iocsexec = (short*)0xa0e;	// IOCSs[N
struct trans_counter trans_counter;			// MJE^
unsigned char rx_buff[2048];				// Mobt@
int imr;									// 荞݋tO
int scsistop;								// SCSI~tO
int intr_type;								// 荞ݎ(0:V-DISP 1:TimerA)
int poll_interval;							// |[OԊu(ݒ)
int poll_current;							// |[OԊu()
int idle;									// AChJE^

#define POLLING_SLEEP	255	// 4-5s

/************************************************
 *   MACAhX擾ߔs                    *
 ************************************************/
int SCSI_GETMACADDR(unsigned char *mac)
{
	unsigned char cmd[10];
	unsigned char sts;
	unsigned char msg;
	
	if (S_SELECT(scsiid) != 0) {
		return 0;
	}

	cmd[0] = 0x28;
	cmd[1] = 0;
	cmd[2] = 1;
	cmd[3] = 0;
	cmd[4] = 0;
	cmd[5] = 0;
	cmd[6] = 0;
	cmd[7] = 0;
	cmd[8] = 6;
	cmd[9] = 0;
	if (S_CMDOUT(10, cmd) != 0) {
		return 0;
	}

#ifdef USE_DMA
	if (S_DATAIN(6, mac) != 0) {
#else
	if (S_DATAIN_P(6, mac) != 0) {
#endif
		return 0;
	}

	S_STSIN(&sts);
	S_MSGIN(&msg);

	return 1;
}

/************************************************
 *   MACAhXݒ薽ߔs                    *
 ************************************************/
int SCSI_SETMACADDR(const unsigned char *mac)
{
	unsigned char cmd[10];
	unsigned char sts;
	unsigned char msg;
	
	if (S_SELECT(scsiid) != 0) {
		return 0;
	}

	cmd[0] = 0x2a;
	cmd[1] = 0;
	cmd[2] = 1;
	cmd[3] = 0;
	cmd[4] = 0;
	cmd[5] = 0;
	cmd[6] = 0;
	cmd[7] = 0;
	cmd[8] = 6;
	cmd[9] = 0;
	if (S_CMDOUT(10, cmd) != 0) {
		return 0;
	}

#ifdef USE_DMA
	S_DATAOUT(6, mac);
#else
	S_DATAOUT_P(6, mac);
#endif

	S_STSIN(&sts);
	S_MSGIN(&msg);
	
	return 1;
}

/************************************************
 *   pPbgMTCY擾ߔs             *
 ************************************************/
int SCSI_GETPACKETLEN(int *len)
{
	unsigned char cmd[10];
	unsigned char buf[2];
	unsigned char sts;
	unsigned char msg;
	
	if (S_SELECT(scsiid) != 0) {
		return 0;
	}

	cmd[0] = 0x28;
	cmd[1] = 0;
	cmd[2] = 1;
	cmd[3] = 1;
	cmd[4] = 0;
	cmd[5] = 0;
	cmd[6] = 0;
	cmd[7] = 0;
	cmd[8] = 2;
	cmd[9] = 0;
	if (S_CMDOUT(10, cmd) != 0) {
		return 0;
	}

#ifdef USE_DMA
	if (S_DATAIN(2, buf) != 0) {
#else
	if (S_DATAIN_P(2, buf) != 0) {
#endif
		return 0;
	}

	S_STSIN(&sts);
	S_MSGIN(&msg);

	*len = (int)(buf[0] << 8) + (int)(buf[1]);

	return 1;
}

/************************************************
 *   pPbgMߔs                       *
 ************************************************/
int SCSI_GETPACKETBUF(unsigned char *buf, int len)
{
	unsigned char cmd[10];
	unsigned char sts;
	unsigned char msg;
	
	if (S_SELECT(scsiid) != 0) {
		return 0;
	}

	cmd[0] = 0x28;
	cmd[1] = 0;
	cmd[2] = 1;
	cmd[3] = 1;
	cmd[4] = 0;
	cmd[5] = 0;
	cmd[6] = (unsigned char)(len >> 16);
	cmd[7] = (unsigned char)(len >> 8);
	cmd[8] = (unsigned char)len;
	cmd[9] = 1;
	if (S_CMDOUT(10, cmd) != 0) {
		return 0;
	}

#ifdef USE_DMA
	if (S_DATAIN(len, buf) != 0) {
#else
	if (S_DATAIN_P(len, buf) != 0) {
#endif
		return 0;
	}

	S_STSIN(&sts);
	S_MSGIN(&msg);

	return 1;
}

/************************************************
 *   pPbgMߔs                       *
 ************************************************/
int SCSI_SENDPACKET(const unsigned char *buf, int len)
{
	unsigned char cmd[10];
	unsigned char sts;
	unsigned char msg;
	
	if (S_SELECT(scsiid) != 0) {
		return 0;
	}

	cmd[0] = 0x2a;
	cmd[1] = 0;
	cmd[2] = 1;
	cmd[3] = 1;
	cmd[4] = 0;
	cmd[5] = 0;
	cmd[6] = (unsigned char)(len >> 16);
	cmd[7] = (unsigned char)(len >> 8);
	cmd[8] = (unsigned char)len;
	cmd[9] = 0;
	if (S_CMDOUT(10, cmd) != 0) {
		return 0;
	}

#ifdef USE_DMA
	S_DATAOUT(len, buf);
#else
	S_DATAOUT_P(len, buf);
#endif
	S_STSIN(&sts);
	S_MSGIN(&msg);
	
	return 1;
}

/************************************************
 *   MACAhX擾                            *
 ************************************************/
int GetMacAddr(struct eaddr* buf)
{
	if (SCSI_GETMACADDR(buf->eaddr) != 0) {
		return 0;
	} else {
		return -1;
	}
}

/************************************************
 *   MACAhXݒ                            *
 ************************************************/
int SetMacAddr(const struct eaddr* data)
{
	if (SCSI_SETMACADDR(data->eaddr) != 0) {
		return 0;
	} else {
		return -1;
	}
}

/************************************************
 *  RaSCSI                                  *
 ************************************************/
int SearchRaSCSI()
{
	int i;
	INQUIRYOPT_T inq;

	for (i = 0; i <= 7; i++) {
		// BRIDGEfoCX
		if (S_INQUIRY(sizeof(INQUIRY_T) , i, (struct INQUIRY*)&inq) < 0) {
			continue;
		}

		if (memcmp(&(inq.info.ProductID), "RASCSI BRIDGE", 13) != 0) {
			continue;
		}

		// TAPԂ擾
		if (S_INQUIRY(sizeof(INQUIRYOPT_T) , i, (struct INQUIRY*)&inq) < 0) {
			continue;
		}
		
		if (inq.options[1] != '1') {
			continue;
		}

		// SCSI IDm
		scsiid = i;
		return 0;
	}

	return -1;
}

/************************************************
 *  RaSCSI ֐                           *
 ************************************************/
int InitRaSCSI(void)
{
#ifdef MULTICAST
	unsigned char multicast_table[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
#endif

	imr = 0;
	scsistop = 0;
	poll_current = -1;
	idle = 0;
	
	return 0;
}

/************************************************
 *  RaSCSI 荞ݏ ֐(|[O)        *
 ************************************************/
void interrupt IntProcess(void)
{
	unsigned char phase;
	unsigned int len;
	int type;
	int_handler func;
	int i;

	// V-DISP GPIP荞݂̓AChJE^Ő
	if (intr_type == 0) {
		// AChZ
		idle++;
		
		// ̏\ɓBĂȂȂXLbv
		if (idle < poll_current) {
			return;
		}

		// AChJE^NA
		idle = 0;
	}
	
	// 荞݊Jn

	// 荞݋̎
	if (imr == 0) {
		return;
	}

	// IOCSsȂ΃XLbv
	if (*iocsexec != -1) {
		return;
	}

	// M͊荞݋֎~
	DI ();

	// oXt[̎
	phase = S_PHASE();
	if (phase != 0) {
		// I
		goto ei_exit;
	}	

	// M
	if (SCSI_GETPACKETLEN(&len) == 0) {
		// RaSCSI~
		scsistop = 1;

		// |[OԊu̍Đݒ(Q)
		UpdateIntProcess(POLLING_SLEEP);

		// I
		goto ei_exit;
	}
	
	// RaSCSI͓쒆
	if (scsistop) {
		scsistop = 0;

		// |[OԊu̍Đݒ(})
		UpdateIntProcess(poll_interval);
	}
	
	// pPbg͓ĂȂ
	if (len == 0) {
		// I
		goto ei_exit;
	}

	// Mobt@փpPbg]
	if (SCSI_GETPACKETBUF(rx_buff, len) == 0) {
		// s
		goto ei_exit;
	}

	// 荞݋
	EI ();
	
	// pPbg^CvŃf[^
	type = rx_buff[12] * 256 + rx_buff[13];
	i = 0;
	while ((func = SearchList2(type, i))) {
		i++;
		func(len, (void*)&rx_buff, ID_EN0);
	}
	trans_counter.recv_byte += len;
	return;

ei_exit:
	// 荞݋
	EI ();
}

/************************************************
 *  RaSCSI pPbgM ֐ (ne.s)         *
 ************************************************/
int SendPacket(int len, const unsigned char* data)
{
	if (len < 1) {
		return 0;
	}
	
	if (len > 1514)	{		// 6 + 6 + 2 + 1500
		return -1;			// G[
	}
	
	// RaSCSI~̂悤ȂG[
	if (scsistop) {
		return -1;
	}

	// M͊荞݋֎~
	DI ();

	// MƑMtOAbv
	if (SCSI_SENDPACKET(data, len) == 0) {
		// 荞݋
		EI ();
		return -1;
	}

	// 荞݋
	EI ();

	// M˗ς
	trans_counter.send_byte += len;
	
	return 0;
}

/************************************************
 *  RaSCSI 荞݋Asݒ ֐        *
 ************************************************/
int SetPacketReception(int i)
{
	imr = i;
	return 0;
}

/************************************************
 * RaSCSI 荞ݏo^                      *
 ************************************************/
void RegisterIntProcess(int n)
{
	volatile unsigned char *p;

	// |[OԊu()̍XVƃAChJE^NA
	poll_current = n;
	idle = 0;
	
	if (intr_type == 0) {
		// V-DISP GPIP荞݃xN^
		// 荞݂Lɂ
		B_INTVCS(0x46, (int)IntProcess);
		p = (unsigned char *)MFP_AEB;
		*p = *p | 0x10;
		p = (unsigned char *)MFP_IERB;
		*p = *p | 0x40;
		p = (unsigned char *)MFP_IMRB;
		*p = *p | 0x40;
	} else if (intr_type == 1) {
		// TimerA̓JEg[hݒ
		VDISPST(NULL, 0, 0);
		VDISPST(IntProcess, 0, poll_current);
	}
}

/************************************************
 * RaSCSI 荞ݏύX                      *
 ************************************************/
void UpdateIntProcess(int n)
{
	// |[OԊu()ƓȂXVȂ
	if (n == poll_current) {
		return;
	}
	
	// |[OԊu()̍XVƃAChJE^NA
	poll_current = n;
	idle = 0;
	
	if (intr_type == 1) {
		// TimerA͍ēo^Kv
		VDISPST(NULL, 0, 0);
		VDISPST(IntProcess, 0, poll_current);
	}
}

// EOF
